<?php

namespace Drupal\update_premium;

use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Extension\ThemeHandlerInterface;
use Drupal\Core\State\StateInterface;
use Drupal\Core\Render\RendererInterface;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\RequestException;
use Symfony\Component\Yaml\Yaml;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Component\Utility\Html;

/**
 * Default implementation of UpdateManagerInterface.
 */
class UpdateManager implements UpdateManagerInterface {
  use DependencySerializationTrait;
  use StringTranslationTrait;

  protected $cache_bin = 'update_premium:projects';

  /**
   * The update settings
   *
   * @var \Drupal\Core\Config\Config
   */
  protected $updateSettings;

  /**
   * An array of installed and enabled projects.
   *
   * @var array
   */
  protected $projects;

  /**
   * The cache backend.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cache;

  /**
   * The HTTP client to fetch the feed data with.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The theme handler.
   *
   * @var \Drupal\Core\Extension\ThemeHandlerInterface
   */
  protected $themeHandler;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $stateStore;

  /**
   * The renderer service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * Constructs a UpdateManager.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache backend.
   * @param \GuzzleHttp\ClientInterface $http_client
   *   A Guzzle client object.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler
   * @param \Drupal\Core\Extension\ThemeHandlerInterface $theme_handler
   *   The theme handler.
   * @param \Drupal\Core\State\StateInterface $state_store
   *   The state service.
   * @param \Drupal\Core\Render\RendererInterface $renderer
   *   The renderer service.
   */
  public function __construct(ConfigFactoryInterface $config_factory, CacheBackendInterface $cache, ClientInterface $http_client, ModuleHandlerInterface $module_handler, ThemeHandlerInterface $theme_handler, StateInterface $state_store, RendererInterface $renderer) {
    $this->updateSettings = $config_factory->get('update.settings');
    $this->httpClient = $http_client;
    $this->moduleHandler = $module_handler;
    $this->themeHandler = $theme_handler;
    $this->stateStore = $state_store;
    $this->cache = $cache;
    $this->renderer = $renderer;
    $this->projects = [];
  }

  /**
   * {@inheritdoc}
   */
  public function checkUpdates($manual = FALSE) {
    $projects = $this->getProjects();
    $projects_hash = md5(serialize($projects));
    // This can be in the middle of a long-running batch, so REQUEST_TIME won't
    // necessarily be valid.
    $request_time_difference = time() - REQUEST_TIME;
    $frequency = $this->updateSettings->get('check.interval_days');
    $interval = 60 * 60 * 24 * $frequency;
    $last_check = $this->stateStore->get('update_premium.last_check') ?: 0;
    $cached_hash = $this->stateStore->get('update_premium.projects_hash') ?: 0;
    if ($cached_hash != $projects_hash) {
      $manual = TRUE;
      $this->stateStore->set('update_premium.projects_hash', $projects_hash);
    }

    if ($manual) {
      $this->cache->delete($this->cache_bin);
      $this->processUpdates($projects);
      Cache::invalidateTags(['update_premium_report']);
      $this->stateStore->set('update_premium.last_check', REQUEST_TIME + $request_time_difference);
    }
    else {
      if ((REQUEST_TIME - $last_check) > $interval) {
        $this->cache->delete($this->cache_bin);
        $this->processUpdates($projects);
        Cache::invalidateTags(['update_premium_report']);
        $this->stateStore->set('update_premium.last_check', REQUEST_TIME + $request_time_difference);
      }
      else {
        $this->processUpdates($projects);
      }
    }
    if ($cache = $this->cache->get($this->cache_bin)) {
      $projects = $cache->data;
    }
    else {
      $this->cache->set($this->cache_bin, $projects);
    }
    return $projects;
  }

  /**
   * {@inheritdoc}
   */
  public function processUpdates(&$projects) {
    if (!empty($projects['modules'])) {
      foreach ($projects['modules'] as $module => $info) {
        $remote = $this->fetchProject($info['update_premium']['update_server']);
        $this->cleanupData($remote);
        if ($remote['version'] != $info['version'] && $remote['machine_name'] == $module) {
          $projects['modules'][$module]['status'] = $this->updateStatus($remote);
          $projects['modules'][$module]['status_icon'] = $this->updateStatusIcon($projects['modules'][$module]['status']);
          $projects['modules'][$module]['update_available'] = $remote;
        }
        else {
          $projects['modules'][$module]['status'] = 'success';
          $projects['modules'][$module]['status_icon'] = $this->updateStatusIcon('success');
        }
      }
    }
    if (!empty($projects['themes'])) {
      foreach ($projects['themes'] as $module => $info) {
        $remote = $this->fetchProject($info['update_premium']['update_server']);
        $this->cleanupData($remote);
        if ($remote['version'] != $info['version'] && $remote['machine_name'] == $module) {
          $projects['themes'][$module]['status'] = $this->updateStatus($remote);
          $projects['themes'][$module]['status_icon'] = $this->updateStatusIcon($projects['themes'][$module]['status']);
          $projects['themes'][$module]['update_available'] = $remote;
        }
        else {
          $projects['themes'][$module]['status'] = 'success';
          $projects['themes'][$module]['status_icon'] = $this->updateStatusIcon('success');
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function fetchProject($url) {
    $data = [];
    try {
      $data = (string) $this->httpClient->get($url, ['headers' => ['Accept' => 'application/json']])->getBody();
    }
    catch (RequestException $exception) {
      watchdog_exception('update', $exception);
    }
    return json_decode($data, TRUE);
  }

  /**
   * {@inheritdoc}
   */
  public function getProjects() {
    foreach ($this->moduleHandler->getModuleList() as $info) {
      if ($data = $this->getProject($info)) {
        $this->projects['modules'][] = $data;
      }
    }
    foreach ($this->themeHandler->listInfo() as $info) {
      if ($data = $this->getProject($info)) {
        $this->projects['themes'][] = $data;
      }
    }
    return $this->projects;
  }

  /**
   * {@inheritdoc}
   */
  public function getProject($info) {
    $data = Yaml::parse(file_get_contents(DRUPAL_ROOT . '/' . $info->getPathname()));
    if (!empty($data['update_premium']) && !empty($data['update_premium']['update_server']) && filter_var($data['update_premium']['update_server'], FILTER_VALIDATE_URL)) {
      $data['update_available'] = [];
      return $data;
    }
  }

  /**
   * Check update status.
   */
  protected function updateStatus($project) {
    $status = 'success';
    if (!empty($project)) {
      if (!empty($project['security_update'])) {
        $status = 'error';
      }
      else {
        $status = 'warning';
      }
    }
    return $status;
  }

  /**
   * Rendered status icon.
   */
  protected function updateStatusIcon($status) {
    switch ($status) {
      case 'error':
        $uri = 'core/misc/icons/e32700/error.svg';
        break;
      case 'warning':
        $uri = 'core/misc/icons/e29700/warning.svg';
        break;
      default:
        $uri = 'core/misc/icons/73b355/check.svg';
        break;
    }
    $icon = [
      '#theme' => 'image',
      '#width' => 18,
      '#height' => 18,
      '#uri' => $uri,
    ];
    return $this->renderer->render($icon);
  }

  /**
   * Cleanup remote data.
   */
  protected function cleanupData(&$data) {
    if (!empty($data) && is_array($data)) {
      foreach ($data as $key => &$value) {
        if (is_array($value)) {
          $this->cleanupData($value);
        }
        else {
          $data[Html::escape($key)] = Html::escape($value);
        }
      }
    }
  }

}
